from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from http.client import HTTPResponse
from io import BytesIO
import json
from pathlib import Path
from typing import List, Union, Dict, TYPE_CHECKING
from urllib import request

import polars as pl
import pyarrow.parquet as pq

if TYPE_CHECKING:
    try:
        import boto3
        from botocore.exceptions import ClientError
    except:
        raise ModuleNotFoundError("Did you install boto3?")

from boomba.core.db import DBManager
from boomba.core.schema import Schema
from boomba.core.config import Config, Conf
from boomba.core.constants import (
    PIPELINE_DIR,
    DATA_DIR,
    ACCEPT_HEADERS,
    CONTENT_HEADERS,
    ENCODING,
    BASE_DATA_FORMAT
)
from boomba.exception.exc import (
    ModuleLocationError,
    PipeNotFoundError,
    UndefinedAttributeError,
    ConfigurationError,
    AttrAccessError
)
from boomba.util.parse import to_snake_case, to_date_format


__all__ = [
    'DBExtractor',
    'APIExtractor'
    'Transformer',
    'FSLoader',
    'DBLoader'
]


class Extractor(ABC):
    data: pl.DataFrame
    schema: Schema
    
    def __init__(self) -> None:
        self._check_attr()
        self._additional_init()
        self.data = self._extract_data()
    
    def _additional_init(self):
        pass

    @abstractmethod
    def _check_attr(self) -> None: ...
    
    @abstractmethod
    def _extract_data(self) -> pl.DataFrame: ...
    
    def _to_dataframe(self, df: pl.DataFrame) -> pl.DataFrame:
        if hasattr(self, 'schema'):
            return pl.DataFrame(df, schema=self.schema().schema)
        return pl.DataFrame(df)
    

class DBExtractor(Extractor):
    """
    A base class for extracting data from a database and returning it as a DataFrame.

    This class provides the core functionality for extracting data from a database using 
    a configured query and parameters, and returning the result as a `pl.DataFrame`. It 
    expects a schema and a query to be provided and requires a valid database configuration 
    to be initialized.
    
    Attributes
    ---
        db_name (str): The database configuration section name.
        schema (Schema): The schema for the data being extracted.
        query (str): The SQL query to be executed.
        query_params (str): The parameters for the query.
        data (pl.DataFrame): The extracted data from database

    """
    db_name: str
    query: str
    query_params: str = None
    _db: DBManager
    
    def _check_attr(self) -> None:
        db_name = 'db_name'
        query = 'query'
        schema = 'schema'
        
        if not hasattr(self, db_name):
            raise UndefinedAttributeError(self, db_name)
            
        if not hasattr(self, query):
            raise UndefinedAttributeError(self, query)
        
        if hasattr(self, schema):
            if not issubclass(self.schema, Schema):
                raise TypeError(
                    ("Invalid type for 'schema'. "
                    "Expected subclass of Schema.")
                )
        
        if hasattr(self, '_db'):
            if not isinstance(self._db, DBManager):
                raise TypeError(
                    "Only DBManager type can set '_db'. "
                    f"class : {self.__class__.__name__}"
                )
    
    def _additional_init(self) -> None:
        if not hasattr(self, '_db'):
            self._db = DBManager(self.db_name)
     
    def _extract_data(self) -> pl.DataFrame:
        try:
            result = self._db.select(self.query, self.query_params, False)
            return self._to_dataframe(result)
        except Exception as e:
            raise RuntimeError(str(e))


class FileExtractor(Extractor):
    pipe_name: str
    collection: str
    start: str
    end: str
    file_name: str
    format: str
    
    def _check_attr(self) -> None:
        ...
    
    def _extract_data(self) -> pl.DataFrame:
        ...


class APIExtractor(Extractor):
    url: str
    method: str
    payload: Dict[str, str]
    
    def _check_attr(self) -> None:
        url = 'url'
        method = 'method'
        payload = 'payload'
        
        if not hasattr(self, url):
            raise UndefinedAttributeError(self, url)
        
        if hasattr(self, method):
            self.method = self.method.upper()
        else:
            raise UndefinedAttributeError(self, method)
            
        if self.method not in ['GET', 'POST']:
            raise ValueError(
                "http method must be one of 'get', or 'post'."
            )
        
        if self.method == 'POST':
            if not hasattr(self, payload):
                raise UndefinedAttributeError(self, payload)
    
    def _parse_json(self, res: HTTPResponse) -> pl.DataFrame:
        try:
            data = res.read().decode()
            return pl.DataFrame(json.loads(data))
        except Exception as e:
            raise RuntimeError(str(e))
    
    def _get(self) -> pl.DataFrame:
        req = request.Request(
            self.url,
            headers=ACCEPT_HEADERS,
            method=self.method
        )
        with request.urlopen(req) as res:
            return self._to_dataframe(self._parse_json(res))
    
    def _post(self) -> pl.DataFrame:
        data = json.dumps(self.payload).encode(ENCODING)
        req = request.Request(
            self.url,
            data=data,
            headers=CONTENT_HEADERS,
            method=self.method
        )
        with request.urlopen(req) as res:
            return self._to_dataframe(self._parse_json(res))

    def _extract_data(self) -> pl.DataFrame:
        if self.method == 'GET':
            return self._get()
        else:
            return self._post()
    

class Transformer:
    """
    Abstract base class for extracting and processing data.

    The `Transformer` class provides the structure for implementing data extraction 
    and processing workflows. It defines an abstract method `process_data()` that 
    must be implemented in subclasses. Data is fetched using the `extractor` attribute, 
    processed via the `process_data()` method, and validated to ensure correctness.

    Attributes
    ----------
    extractor : Union[List[DBExtractor], DBExtractor]
        The object or list of objects responsible for extracting data.
    data : pl.DataFrame
        The raw data extracted by the extractor(s).
    result : Optional[Union[Dict[str, pl.DataFrame], pl.DataFrame]]
        The processed result after applying `process_data()`.
    """
    extractor: Union[List[DBExtractor], Extractor]
    
    def __init__(self) -> None:
        self.data = self._get_extracted_data()
        self.result = self._validate_process_data()
    
    def _get_extracted_data(self) -> Union[Dict[str, pl.DataFrame], pl.DataFrame]:
        """
        Extracts data using the `extractor` attribute.
        
        Returns
        -------
        Union[Dict[str, pl.DataFrame], pl.DataFrame]
        - a dictionary where keys are the name of subclasses of Extractor class,
        and values are the extracted data as `pl.DataFrame` for each class.
        """
        if not hasattr(self, 'extractor'):
            raise UndefinedAttributeError(self, 'extractor')
        
        data_map = {}
                
        if isinstance(self.extractor, list):
            if not all(issubclass(item, DBExtractor) for item in self.extractor):
                raise TypeError("All items in the list must be of type DBExtractor.")
        
        elif issubclass(self.extractor, Extractor):
            self.extractor = [self.extractor]
        else:
            raise TypeError(
                "extractor must be a List[Extractor] or a Extractor."
            )
        
        try:
            for cls in self.extractor:
                inst = cls()
                data_map[inst.__class__.__name__] = inst.data
        except Exception as e:
            raise RuntimeError(str(e))
        
        return data_map
    
    def _validate_process_data(self) -> pl.DataFrame:
        """
        Calls the `process_data()` method and validates the result.
        This method raises a `TypeError` if the `process_data()` method returns `None`.
        """
        result = self.process_data()
        
        if result is None:
            raise NotImplementedError(
                (f"Class {self.__class__.__name__} must implement "
                 "the 'process_data' method, "
                 "and return a DataFrame.")
            )
        
        if not isinstance(result, pl.DataFrame):
            raise TypeError(
                "The result produced by the 'process_data()' method "
                "must be of type 'pl.DataFrame'."
            )
        
        return result
        
    def process_data(self) -> pl.DataFrame:
        """
        Abstract method. This method must be implemented to process the data.
        
        This method must be implemented in subclasses of `Transformer`.
        If not implemented, a `TypeError` will be raised.
        """
        pass


class Loader(ABC):
    """
    An Abstract class responsible for loading and processing data within a pipeline.
    
    Attributes
    -----
    - schema (Schema, optional): The schema for loading the data.
    - transformer (Transformer, required): The subclass of the transformer class used to process the data.
    - collection (str, optional): Data separator (concept corresponding to a database table).
    - file_name (str, optional): File name.
    - pipe_name (str, read only): Pipe name.
    - location (Path, read only): The directory where the data will be stored, determined by 'pipe_name / collection'.
    - path (str, read only): The data storage path, determined by 'location / file_name'.
    """
    schema: Schema
    transformer: Transformer
    collection: str
    file_name: str
    pipe_name: str
    location: Path
    path: str
    _module_setter_for_test: str # Test use only, no other usage allowed
    
    def __init__(self, conf: Config=Conf) -> None:
        self._check_attr()
        self._conf = conf
        self._additional_init_()
        self.__pipe_name = self._get_pipe_name()
        self.__collection = self._set_collection()
        self.__location = self._set_location()
        self.__path = self._set_path()
        self._load_data()
        
    def __repr__(self) -> str:
        return (
            f"Loader(pipe_name={self.pipe_name}, "
            f"collection={self.collection})"
        )
    
    @property
    def pipe_name(self) -> str:
        return self.__pipe_name
    
    @property
    def collection(self) -> str:
        return self.__collection
    
    @property
    def location(self) -> Path:
        return self.__location
    
    @property
    def path(self) -> str:
        return self.__path

    @abstractmethod
    def _additional_init_(self) -> None: ...
    
    @abstractmethod
    def _set_path(self) -> str: ...
    
    @abstractmethod
    def _load_data(self) -> None: ...
    
    def _check_attr(self) -> None:
        transformer = 'transformer'
        collection = 'collection'
        
        if not hasattr(self, transformer):
            raise UndefinedAttributeError(self, transformer)
        else:
            if not isinstance(self.transformer, type) or not issubclass(self.transformer, Transformer):
                raise TypeError(
                    ("Invalid type for 'transformer'. "
                    "Expected instance of 'Transformer'.")
                )
            
        if hasattr(self, collection):
            if not isinstance(self.collection, str):
                raise TypeError(
                    ("Invalid type for 'collection'. "
                    "Expected instance of 'str'.")
                )
        
    def _get_pipe_name(self) -> str:
        """
        Raises
        -----
            RuntimeError: If the Location class is used outside of a valid pipeline.
        """
        if not hasattr(self, '_module_setter_for_test'):
            token = self.__module__.split('.')
        else:
            token = self._module_setter_for_test.split('.')
        
        if token[0] != PIPELINE_DIR.name:
            raise ModuleLocationError(
                self,
                additional_msg=self.__module__
            )
        
        if len(token) == 1:
            raise PipeNotFoundError(self)
        
        return token[1]
    
    def _set_collection(self) -> str:
        if not hasattr(self, 'collection'):
            return to_snake_case(self.__class__.__name__)
        return self.collection.lower()
    
    def _set_location(self) -> Path:
        return Path(self.pipe_name) / self.collection


class FSType(Enum):
    local = 'local'
    s3 = 's3'


class FSLoader(Loader):
    fs_name: str
    _fs_type: str
    bucket: str
    
    def _additional_init_(self) -> None:
        self.fs_name = self._set_fs_name()
        self._fs_type = self._get_fs_type()
        self.bucket = self._set_bucket()
        self._s3 = self._get_s3_client()
        
    def _set_fs_name(self) -> str:
        if not hasattr(self, 'fs_name'):
            return self._conf.base_fs
        return self.fs_name
    
    def _get_fs_type(self) -> str:
        return self._conf.file_system[self.fs_name]['fs_type']
    
    def _set_bucket(self) -> str:
        if self._fs_type == FSType.local.value:
            if hasattr(self, 'bucket'):
                raise AttributeError(
                    "Not allowed to set bucket on the local system"
                )
        elif self._fs_type == FSType.s3.value:
            if not hasattr(self, 'bucket'):
                self.bucket = self._conf.file_system[self.fs_name]['bucket']
            return self.bucket
        else:
            raise ValueError(f"Invalid 'fs_type'. Not allow {self._fs_type}")
    
    def _get_s3_client(self):
        '''Import boto3 only when using S3 (not a default dependency package).'''
        try:
            global boto3, ClientError
            if self._fs_type == FSType.s3.value:
                import boto3 as boto3_
                from botocore.exceptions import ClientError as ClientError_
                boto3 = boto3_
                ClientError = ClientError_
                
                try:
                    access_id = self._conf.file_system[self.fs_name]['access_id']
                    access_key = self._conf.file_system[self.fs_name]['access_key']
                    region = self._conf.file_system[self.fs_name]['region']
                except:
                    raise ConfigurationError(self, additional_msg='Access info for s3.')
                return boto3.client(
                    FSType.s3.value,
                    aws_access_key_id=access_id,
                    aws_secret_access_key=access_key,
                    region_name=region
                )
        except Exception as e:
            raise RuntimeError(str(e))
    
    def _set_path(self) -> str:
        if not hasattr(self, 'file_name'):
            format = to_date_format(self._conf.file_date_format)
            self.file_name = datetime.now().strftime(format)
        
        self.file_name = f'{self.file_name}.{BASE_DATA_FORMAT}'
        
        if self._fs_type == FSType.local.value:
            dir = DATA_DIR / self.location
            dir.mkdir(parents=True, exist_ok=True)
            path = dir / self.file_name
            
            if not self._conf.file_system[self.fs_name]['allow_overwrite']:
                if path.exists():
                    raise FileExistsError(
                        "Already exsits the same file in the path : "
                        f"'{str(path)}'"
                    )
        elif self._fs_type == FSType.s3.value:
            path = str(self.location / self.file_name).replace('\\', '/')
            if not self._conf.file_system[self.fs_name]['allow_overwrite']:
                try:
                    self._s3.head_object(Bucket=self.bucket, Key=path)
                    raise FileExistsError(
                        "Already exsits the same file in the path : "
                        f"'{str(path)}'"
                    )
                except ClientError as e:
                    if e.response['Error']['Code'] == '404':
                        pass
        
        return str(path)
      
    def _load_data(self) -> None:
        transformer: Transformer = self.transformer()
        data = transformer.result
        
        if hasattr(self, 'schema'):
            data = pl.DataFrame(
                data,
                schema=self.schema().schema
            )
        else:
            data = pl.DataFrame(data)
        
        if self._fs_type == FSType.local.value:
            data.write_parquet(self.path)
        elif self._fs_type == FSType.s3.value:
            buffer = BytesIO()
            pq.write_table(data.to_arrow(), buffer)
            buffer.seek(0)
            self._s3.put_object(
                Bucket=self.bucket,
                Key=self.path,
                Body=buffer.read())
            

class DBLoader(Loader):
    table_name: str
    db_name: str
    _db: DBManager
    
    def _additional_init_(self) -> None:
        if not hasattr(self, 'table_name'):
            raise UndefinedAttributeError(self, additional_msg='table_name')
        
        if not hasattr(self, 'db_name'):
            raise UndefinedAttributeError(self, additional_msg='db_name')
    
    def _set_path(self) -> None:
        return self.table_name
    
    def _load_data(self) -> None:
        if not hasattr(self, '_db'):
            self._db = DBManager(self.db_name)

        transformer: Transformer = self.transformer()
        data = transformer.result
        self._db.insert(data, self.table_name)


from datetime import datetime, date
from typing import List
from typing import Optional, Tuple, Deque

from sqlalchemy import (
    Integer,
    String,
    DateTime,
    Date,
    Boolean,
    ForeignKey,
    UniqueConstraint
)
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    mapped_column,
    relationship,
    sessionmaker,
    Session
)

from boomba.core.config import Conf
from boomba.core.constants import BASE_DATA_FORMAT
from boomba.core.db import Connector
from boomba.core.etl import Loader

__all__ = [
    'Base'
    'JobInfo',
    'ScheduleInfo',
    'DataDirectory',
    'DataPath',
    'SaveMetaData'
]

_connector = Connector(Conf.base_db, Conf)


class Base(DeclarativeBase):
    pass


class JobInfo(Base):
    __tablename__ = "job_info"
    
    id: Mapped[int] = mapped_column(primary_key=True)
    job_name: Mapped[str] = mapped_column(String(32), unique=True)
    registered_at: Mapped[datetime] = mapped_column(DateTime, default=datetime.now)
    start_date: Mapped[Optional[date]] = mapped_column(Date, nullable=True)
    end_date: Mapped[Optional[date]] = mapped_column(Date, nullable=True)
    cron_expression: Mapped[str] = mapped_column(String(100))
    
    schedule_info: Mapped[List["ScheduleInfo"]] = relationship(
        "ScheduleInfo", back_populates="job_info", cascade='save-update'
    )
    data_directory: Mapped["DataDirectory"] = relationship(
        "DataDirectory", back_populates="job_info", uselist=False, cascade='save-update'
    )
    
    def __repr__(self) -> str:
        return f"JobInfo(id={self.id!r}, name={self.job_name!r})"


class ScheduleInfo(Base):
    __tablename__ = "schedule_info"
    
    id: Mapped[int] = mapped_column(primary_key=True)
    job_id: Mapped[int] = mapped_column(ForeignKey("job_info.id"))
    scheduled_at: Mapped[datetime] = mapped_column(DateTime)
    executed_at: Mapped[datetime] = mapped_column(DateTime, default=None, nullable=True)
    is_completed: Mapped[bool] = mapped_column(Boolean, default=False)
    path_id: Mapped[Optional[int]] = mapped_column(ForeignKey("data_path.id"))
    
    job_info: Mapped["JobInfo"] = relationship(
        "JobInfo", back_populates="schedule_info", cascade='save-update'
    )
    data_path: Mapped["DataPath"] = relationship(
        "DataPath", back_populates="schedule_info", uselist=False, cascade='save-update'
    )
    
    __table_args__ = (
        UniqueConstraint('job_id', 'scheduled_at', name='_job_schedule_uc'),
    )
    
    def __repr__(self) -> str:
        return (
            f"ScheduleInfo(id={self.id!r}, "
            f"job_name={self.job_id!r}), "
            f"scheduled_at={self.scheduled_at!r}, "
            f"executed_at={self.executed_at})"
        )

  
class DataDirectory(Base):
    __tablename__ = 'data_directory'
    
    id: Mapped[int] = mapped_column(primary_key=True)
    pipe_name: Mapped[str] = mapped_column(String(32))
    collection: Mapped[str] = mapped_column(String(32))
    created_at: Mapped[datetime] = mapped_column(DateTime, default=datetime.now)
    job_id: Mapped[Optional[int]] = mapped_column(ForeignKey("job_info.id"))
    
    data_path: Mapped[List["DataPath"]] = relationship(
        "DataPath", back_populates="data_directory", cascade='save-update'
    )
    job_info: Mapped["JobInfo"] = relationship(
        "JobInfo", back_populates="data_directory", uselist=False, cascade='save-update'
    )
    
    __table_args__ = (
        UniqueConstraint('pipe_name', 'collection', name='_pipe_collection_uc'),
    )
    
    def __repr__(self) -> str:
        return (
            f"DataPath(id={self.id!r}, "
            f"file_name={self.pipe_name!r}, "
            f"file_type={self.collection!r})"
        )


class DataPath(Base):
    __tablename__ = 'data_path'
    
    id: Mapped[int] = mapped_column(primary_key=True)
    dir_id: Mapped[int] = mapped_column(ForeignKey('data_directory.id'))
    file_name: Mapped[str] = mapped_column(String(32))
    file_type: Mapped[str] = mapped_column(String(32))
    update_count: Mapped[int] = mapped_column(Integer, default=0)
    created_at: Mapped[datetime] = mapped_column(DateTime, default=datetime.now)
    updated_at: Mapped[datetime] = mapped_column(DateTime, default=datetime.now)
    
    data_directory: Mapped["DataDirectory"] = relationship(
        "DataDirectory", back_populates="data_path", cascade='save-update'
    )
    schedule_info: Mapped[List["ScheduleInfo"]] = relationship(
        "ScheduleInfo", back_populates="data_path", cascade='save-update', uselist=True
    )
    
    __table_args__ = (
        UniqueConstraint('dir_id', 'file_name', name='_dir_file_uc'),
    )
    
    def __repr__(self) -> str:
        return (
            f"DataPath(id={self.id!r}, "
            f"file_name={self.file_name!r}, "
            f"file_type={self.file_type!r})"
        )


class QueueInfo(Base):
    __tablename__ = 'queue_info'
    
    queue_no: Mapped[int] = mapped_column(primary_key=True)
    job_name: Mapped[str] = mapped_column(String(32))
    scheduled_at: Mapped[datetime] = mapped_column(DateTime)
    
    def __repr__(self) -> str:
        return (
            f"QueueInfo(queue_no={self.queue_no}, "
            f"job_name={self.job_name}, scheduled_at={self.scheduled_at})"
        )


class MetaDataManager:
    
    def __init__(self, connector: Connector=_connector) -> None:
        self._connector = connector
    
    def _save_session(func):
        def wrapper(self, *args, **kwargs):
            try:
                engine = self._connector.engine
                with sessionmaker(engine)() as session:
                    obj = func(self, session, *args, **kwargs)
                    session.commit()
                    return obj
                
            except Exception as e:
                if session:
                    session.rollback()
                raise RuntimeError(str(e))
        return wrapper
    
    @_save_session
    def init_schdule(
        self,
        session: Session,
        datetime: datetime
    ) -> None:
        session.query(ScheduleInfo)\
               .filter(ScheduleInfo.scheduled_at > datetime)\
               .delete()
    
    @_save_session
    def save_job(
        self,
        session: Session,
        job_name: str,
        registered_at: datetime,
        cron_expression: str,
        start_date: str=None,
        end_date: str=None
    ) -> JobInfo:
        job = session.query(JobInfo)\
                     .filter_by(job_name=job_name)\
                     .first()
        
        if job is None:
            if start_date is None:
                start_date = datetime.now()
                
            job = JobInfo(
                job_name=job_name,
                registered_at=registered_at,
                start_date=start_date,
                end_date=end_date,
                cron_expression=cron_expression
            )
            session.add(job)
        
        return job
    
    @_save_session
    def save_schedule(
        self,
        session: Session,
        scheduled_at: datetime,
        job_name: str,
        executed_at: datetime=None,
        is_completed: bool=None,
    ) -> ScheduleInfo:
        job_info = session.query(JobInfo)\
                          .filter_by(job_name=job_name).first()
        schedule = session.query(ScheduleInfo)\
                          .filter_by(job_id=job_info.id, scheduled_at=scheduled_at)\
                          .first()
        
        if schedule is None:
            schedule = ScheduleInfo(
                scheduled_at=scheduled_at,
                job_info=job_info
            )
            session.add(schedule)
        else:
            schedule.is_completed = is_completed
            schedule.executed_at = executed_at
            session.add(schedule)
        
        return schedule
    
    @_save_session    
    def save_dir(
        self,
        session: Session,
        pipe_name: str,
        collection: str,
        job_name: str
    ) -> DataDirectory:
        dir = session.query(DataDirectory)\
                     .filter_by(pipe_name=pipe_name, collection=collection)\
                     .first()
        
        if dir is None:
            job_info = session.query(JobInfo)\
                              .filter_by(job_name=job_name).first()
            dir = DataDirectory(
                pipe_name=pipe_name,
                collection=collection,
                job_info=job_info
            )
            session.add(dir)
        
        return dir
    
    @_save_session
    def save_path(
        self,
        session: Session,
        file_name: str,
        pipe_name: str,
        collection: str,
        scheduled_at: str,
        file_type: str=None
    ) -> DataPath:
        data_dir: DataDirectory = session.query(DataDirectory)\
                          .filter_by(pipe_name=pipe_name, collection=collection)\
                          .first()
        schedule_info = session.query(ScheduleInfo)\
                               .filter_by(job_id=data_dir.job_id, scheduled_at=scheduled_at)\
                               .first()
        path = session.query(DataPath)\
                      .filter_by(dir_id=data_dir.id, file_name=file_name)\
                      .first()
        
        if path:
           path.updated_at = datetime.now()
           path.update_count += 1
           path.schedule_info.append(schedule_info)
        else:
            if file_type is None:
                file_type = BASE_DATA_FORMAT
            
            path = DataPath(
                file_name=file_name,
                file_type=file_type,
                dir_id=data_dir.id,
                schedule_info=[schedule_info]
            )
        
        session.add(path)
        return path
    
    @_save_session
    def save_queue(
        self,
        session: Session,
        job_queue: Deque[Tuple[Loader, datetime, str]]
    ) -> None:
        '''
        Args
        ----
        - job_list(str, str) : a list of a pair of 'scheduled_at', and 'job_name'
        '''
        session.query(QueueInfo).delete()
        
        for i, (_, scheduled_at, job_name) in enumerate(job_queue):
            queue = QueueInfo(queue_no=i+1, scheduled_at=scheduled_at, job_name=job_name)
            session.add(queue)


from pathlib import Path

import unittest

from tempfile import NamedTemporaryFile
from boomba.core.config import Config
from boomba.core.constants import MOCK_CONFIG


class TestConfigFileReader(unittest.TestCase):
    
    def test_load_non_empty(self):
        content = 'FIRST_VAR="first_value"; SECOND_VAR="second_value"'
        return_value = {'FIRST_VAR': 'first_value', 'SECOND_VAR': 'second_value'}

        with NamedTemporaryFile(mode='w+t', delete=False) as f:
            f.write(content)
            tmp_path = Path(f.name)
        
        try:
            config = Config(tmp_path)
            self.assertEqual(config._vars, return_value)
        finally:
            tmp_path.unlink()
    
    def test_load_empty(self):
        # required non-exist-path
        path = Path('not a path!')
        if path.exists():
            raise Exception(f"path exists. {path}")
        
        config = Config(path)
        self.assertEqual(config._vars, MOCK_CONFIG)
    
    def test_set_attr(self):
        first_flag = False
        second_flag = False

        content = 'FIRST_VAR="first_value"; SECOND_VAR="second_value"'
        
        with NamedTemporaryFile(mode='w+t', delete=False) as f:
            f.write(content)
            tmp_path = Path(f.name)
        
        try:
            config = Config(tmp_path)

            if config.first_var == 'first_value':
                first_flag = True
            
            if config.second_var == 'second_value':
                second_flag = True
            
            self.assertTrue(all([first_flag, second_flag]))
        finally:
            tmp_path.unlink()


import sys
import unittest
from unittest.mock import MagicMock

from boomba.cli.command import CommandRegistor
from boomba.cli.description import HELP


class TestRegistor(unittest.TestCase):

    def setUp(self):
        sys.argv = ['startproject']
        self.inst = CommandRegistor()
    
    def test_add_startproject(self):
        self.inst._subparser = MagicMock()
        self.inst.add_startproject()
        self.inst._subparser.add_parser.assert_called_with(
            self.inst.startproject,
            help=HELP[self.inst.startproject]
        )
    
    def test_add_initdb(self):
        self.inst._subparser = MagicMock()
        self.inst.add_initdb()
        self.inst._subparser.add_parser.assert_called_with(
            self.inst.initdb,
            help=HELP[self.inst.initdb]
        )
    
    def test_add_run(self):
        self.inst._subparser = MagicMock()
        self.inst.add_run()
        self.inst._subparser.add_parser.assert_called_with(
            self.inst.run,
            help=HELP[self.inst.run]
        )
    
    def test_add_createpipe(self):
        parser_flag = False
        option_flag = False
        
        if self.inst.createpipe in self.inst._subparser.choices:
            parser_flag = True
        
            subparser = self.inst._subparser.choices[self.inst.createpipe]
            for action in subparser._actions:
                print(action.dest)
                if action.dest == 'name':
                    option_flag = True
                    break
        self.assertTrue(all([parser_flag, option_flag]))
    
    def test_add_test(self):
        parser_flag = False
        option_flag = False
        
        if self.inst.test in self.inst._subparser.choices:
            parser_flag = True
        
            subparser = self.inst._subparser.choices[self.inst.test]
            for action in subparser._actions:
                print(action.dest)
                if action.dest == 'name':
                    option_flag = True
                    break
        self.assertTrue(all([parser_flag, option_flag]))




class BaseExc(Exception):
    default_msg: str
    
    def __init__(
        self,
        instance: object=None,
        msg: str=None,
        additional_msg: str=None
    ) -> None:
        if additional_msg:
            self.default_msg += additional_msg
            
        msg = msg or self.default_msg
        
        if instance:
            msg = f"{msg} (class: {instance.__class__.__name__})"
        
        super().__init__(msg)


class NonEmptyDirectoryError(BaseExc):
    """Raised when the target directory for project initialization is not empty."""
    default_msg = 'Root directory must be empty'


class ConfigurationError(BaseExc):
    default_msg = 'The configuration is missing. '

class DatabaseConfigurationError(BaseExc):
    default_msg = (
        "Base database failed to connect. "
        "Please check your database or configuration."
    )
    
class EmptyArgumentError(BaseExc):
    default_msg = (
        "No arguments were provided. Please refer to the help.\n"
        " $ boomba --help"
    )
    
class LogConfigurationError(BaseExc):
    default_msg = "The log configuration has not been set."
    
class DirectoryNotFoundError(BaseExc):
    default_msg = "There is no such directory."
    
class ModuleLocationError(BaseExc):
    default_msg = (
        "The Location class can only be used "
        "within the apps of the pipeline. "
        "The current usage location is inappropriate. "
        "module Location : "
    )

class PipeNotFoundError(BaseExc):
    default_msg = (
        "The Load class requires an app "
        "to be registered in the pipeline before use."
    )
    
class UndefinedMethodError(BaseExc):
    default_msg = "Method not implemented. method="
    
class UndefinedAttributeError(BaseExc):
    default_msg = "Missing required attribute: "
    
class LogDirectoryError(BaseExc):
    default_msg = "Not allowed to use 'system'"

class DBConnectionError(BaseExc):
    default_msg = "Invaild database configuration. "

class AttrAccessError(BaseExc):
    default_msg = "Attempted to access an invalid attribute: "


from abc import ABC, abstractmethod
from typing import Union
import polars as pl
import pyarrow as pa

__all__ = ['Dtype']

class _Dtypes(ABC):
    @property
    @abstractmethod
    def dtype(self) -> Union[pa.DataType, pl.DataType]:
        pass

# Datatypes for Polars
class Dtype:
    
    class Int8(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.Int8
        
    class Int16(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.Int16

    class Int32(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.Int32

    class Int64(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.Int64

    class UInt8(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.UInt8

    class UInt16(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.UInt16

    class UInt32(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.UInt32

    class UInt64(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.UInt64

    class Float32(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.Float32

    class Float64(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.Float64

    class String(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.Utf8

    class Boolean(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.Boolean

    class Date(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.Date

    class Time(_Dtypes):
        @property
        def dtype(self) -> pl.DataType:
            return pl.Time

    class Datetime(_Dtypes):
        '''
        Data type representing a calendar date and time of day.
        
        Parameters
        ---
        time_unit : {'us', 'ns', 'ms'}
            Unit of time. Defaults to 'ms' (milliseconds).
        '''
        def __init__(self, time_unit: str='ms') -> None:
            self.time_unit = time_unit
        
        @property
        def dtype(self) -> pl.DataType:
            return pl.Datetime(self.time_unit)

    class Decimal(_Dtypes):
        def __init__(self, precision: int, scale: int) -> None:
            self.precision = precision
            self.scale = scale

        @property
        def dtype(self) -> pl.DataType:
            return pl.Decimal(self.precision, self.scale)

    class __List(_Dtypes):
        def __init__(self, dtype: _Dtypes) -> None:
            self._dtype = dtype

        @property
        def dtype(self) -> pl.DataType:
            return pl.List(self._dtype.dtype)

    class __Struct(_Dtypes):
        def __init__(self, fields: dict) -> None:
            self.fields = fields

        @property
        def dtype(self) -> pl.DataType:
            return pl.Struct([(k, v.dtype) for k, v in self.fields.items()])


# Datatypes for Arrow
class _ArrowDtypes:

    class Int8(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.int8()

    class Int16(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.int16()

    class Int32(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.int32()

    class Int64(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.int64()

    class UInt8(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.uint8()

    class UInt16(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.uint16()

    class UInt32(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.uint32()

    class UInt64(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.uint64()

    class Float16(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.float16()

    class Float32(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.float32()

    class Float64(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.float64()

    class Boolean(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.bool_()

    class String(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.string()

    class Binary(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.binary()

    class LargeString(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.large_string()

    class LargeBinary(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.large_binary()

    class Date32(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.date32()

    class Date64(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.date64()

    class Timestamp(_Dtypes):
        def __init__(self, unit="ns", tz=None):
            self.unit = unit
            self.tz = tz

        @property
        def dtype(self) -> pa.DataType:
            return pa.timestamp(self.unit, self.tz)

    class Time32(_Dtypes):
        def __init__(self, unit="s"):
            self.unit = unit

        @property
        def dtype(self) -> pa.DataType:
            return pa.time32(self.unit)

    class Time64(_Dtypes):
        def __init__(self, unit="ns"):
            self.unit = unit

        @property
        def dtype(self) -> pa.DataType:
            return pa.time64(self.unit)

    class Decimal128(_Dtypes):
        def __init__(self, precision, scale):
            self.precision = precision
            self.scale = scale

        @property
        def dtype(self) -> pa.DataType:
            return pa.decimal128(self.precision, self.scale)

    class Decimal256(_Dtypes):
        def __init__(self, precision, scale):
            self.precision = precision
            self.scale = scale

        @property
        def dtype(self) -> pa.DataType:
            return pa.decimal256(self.precision, self.scale)

    class ListType(_Dtypes):
        def __init__(self, value_type):
            self.value_type = value_type

        @property
        def dtype(self) -> pa.DataType:
            return pa.list_(self.value_type)

    class LargeListType(_Dtypes):
        def __init__(self, value_type):
            self.value_type = value_type

        @property
        def dtype(self) -> pa.DataType:
            return pa.large_list(self.value_type)

    class MapType(_Dtypes):
        def __init__(self, key_type, item_type):
            self.key_type = key_type
            self.item_type = item_type

        @property
        def dtype(self) -> pa.DataType:
            return pa.map_(self.key_type, self.item_type)

    class StructType(_Dtypes):
        def __init__(self, fields):
            self.fields = fields

        @property
        def dtype(self) -> pa.DataType:
            return pa.struct(self.fields)

    class NullType(_Dtypes):
        @property
        def dtype(self) -> pa.DataType:
            return pa.null()


from datetime import datetime
import unittest
import polars as pl
from boomba.core.dtypes import Dtype

class TestDtypes(unittest.TestCase):

    def test_int8(self):
        dtype_instance = Dtype.Int8()
        data = pl.DataFrame({"value": [1, 2, 3]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Int8)

    def test_int16(self):
        dtype_instance = Dtype.Int16()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Int16)
        
    def test_int32(self):
        dtype_instance = Dtype.Int32()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Int32)
    
    def test_int64(self):
        dtype_instance = Dtype.Int64()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Int64)

    def test_uint8(self):
        dtype_instance = Dtype.UInt8()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.UInt8)
        
    def test_uint8(self):
        dtype_instance = Dtype.UInt8()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.UInt8)
    
    def test_int8(self):
        dtype_instance = Dtype.Int8()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Int8)

    def test_int16(self):
        dtype_instance = Dtype.Int16()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Int16)

    def test_int32(self):
        dtype_instance = Dtype.Int32()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Int32)

    def test_int64(self):
        dtype_instance = Dtype.Int64()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Int64)

    def test_uint8(self):
        dtype_instance = Dtype.UInt8()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.UInt8)

    def test_uint16(self):
        dtype_instance = Dtype.UInt16()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.UInt16)

    def test_uint32(self):
        dtype_instance = Dtype.UInt32()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.UInt32)

    def test_uint64(self):
        dtype_instance = Dtype.UInt64()
        data = pl.DataFrame({"value": [10, 20, 30]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.UInt64)

    def test_float32(self):
        dtype_instance = Dtype.Float32()
        data = pl.DataFrame({"value": [10.5, 20.5, 30.5]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Float32)

    def test_float64(self):
        dtype_instance = Dtype.Float64()
        data = pl.DataFrame({"value": [10.5, 20.5, 30.5]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Float64)

    def test_string(self):
        dtype_instance = Dtype.String()
        data = pl.DataFrame({"value": ["a", "b", "c"]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Utf8)

    def test_boolean(self):
        dtype_instance = Dtype.Boolean()
        data = pl.DataFrame({"value": [True, False, True]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Boolean)

    def test_date(self):
        dtype_instance = Dtype.Date()
        data = pl.DataFrame({"value": ["2022-01-01", "2022-02-01", "2022-03-01"]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Date)

    def test_time(self):
        dtype_instance = Dtype.Time()
        time_data = [
            datetime.strptime(t, "%H:%M:%S").time() for t in ["12:00:00", "14:30:00", "16:45:00"]
        ]
        data = pl.DataFrame({"value": time_data}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Time)

    def test_datetime_millisecond(self):
        dtype_instance = Dtype.Datetime('ms')
        data = pl.DataFrame({
            "value": [
                datetime(2022, 1, 1, 12, 0, 0),
                datetime(2022, 2, 1, 14, 30, 0),
                datetime(2022, 3, 1, 16, 45, 0)
            ]
        }, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Datetime(time_unit='ms'))

    def test_datetime_microsecond(self):
        dtype_instance = Dtype.Datetime('us')
        data = pl.DataFrame({
            "value": [
                datetime(2022, 1, 1, 12, 0, 0),
                datetime(2022, 2, 1, 14, 30, 0),
                datetime(2022, 3, 1, 16, 45, 0)
            ]
        }, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Datetime(time_unit="us"))

    def test_datetime_nanosecond(self):
        dtype_instance = Dtype.Datetime('ns')
        data = pl.DataFrame({
            "value": [
                datetime(2022, 1, 1, 12, 0, 0),
                datetime(2022, 2, 1, 14, 30, 0),
                datetime(2022, 3, 1, 16, 45, 0)
            ]
        }, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Datetime(time_unit="ns"))

    def test_decimal(self):
        dtype_instance = Dtype.Decimal(precision=10, scale=2)
        data = pl.DataFrame({"value": [10.5, 20.5, 30.5]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.Decimal(10, 2))

    def test_list(self):
        '''
        dtype_instance = Dtype.List(Dtype.Int8())
        data = pl.DataFrame({"value": [[10, 20], [30, 40], [50, 60]]}, schema=[("value", dtype_instance.dtype)])
        self.assertEqual(data["value"].dtype, pl.List(pl.Int8))
        '''
        ...

    def test_struct(self):
        '''
        dtype_instance = Dtype.Struct({"field1": Dtype.Int8(), "field2": Dtype.Boolean()})
        data = pl.DataFrame({"field1": [10, 20, 30], "field2": [True, False, True]}, schema=[("field1", dtype_instance.dtype), ("field2", dtype_instance.dtype)])
        self.assertEqual(data["field1"].dtype, pl.Int8)
        self.assertEqual(data["field2"].dtype, pl.Boolean)
        '''
        ...


import pytest
from pathlib import Path
import polars as pl
from tempfile import NamedTemporaryFile

from boomba.core.db import DBManager
from boomba.core.config import Config
from boomba.core.dtypes import Dtype
from boomba.core.etl import (
    Extractor,
    DBExtractor,
    APIExtractor,
    Transformer,
    Loader,
    DBLoader,
    FSLoader
)
from boomba.core.schema import Schema


def test_extractor():
    class TestSchema(Schema):
        name = Dtype.String()
        age = Dtype.UInt8()

    class TestExtractor(Extractor):
        schema = TestSchema
        
        def _check_attr(self):
            pass

        def _extract_data(self):
            return pl.DataFrame({
                "name": ["Alice", "Bob", "Paul"],
                "age": [20.0, 30.0, 40.0]
            })
    
    test = TestExtractor()
    expected_value = [('name', pl.String), ('age', pl.Float64)]
    assert list(test.data.schema.items()) == expected_value
    test.data = test._to_dataframe(test.data)
    expected_value = [('name', pl.String), ('age', pl.UInt8)]
    assert list(test.data.schema.items()) == expected_value


@pytest.fixture
def prepare_test_config():
    
    content = '''
DATABASE = {
    'test': {
        'drivername': 'sqlite',
        'username': None,
        'password': None,
        'host': None,
        'port': None,
        'database': ':memory:'        
    }
}
'''
    with NamedTemporaryFile(mode='w+t', delete=False) as f:
        f.write(content)
        tmp_path = Path(f.name)
    
    yield Config(tmp_path)
    tmp_path.unlink()


def test_dbextractor_sqlite(prepare_test_config):
    db = DBManager('test', prepare_test_config)
    data = pl.DataFrame({
        "name": ["Alice", "Bob", "Paul"],
        "age": [20.0, 30.0, 40.0]
    })
    db.insert(data, 'test_table')

    class TestExtractor(DBExtractor):
        db_name = 'test'
        query = 'SELECT * FROM test_table'
        _db = db
    
    extractor = TestExtractor()
    assert extractor.data.equals(data)


def test_apiextractor():
    class TestAPI(APIExtractor):
        url = "https://jsonplaceholder.typicode.com/posts/1"
        method = 'GET'

    test = TestAPI()
    expect = {
        "userId": 1,
        "id": 1,
        "title": "sunt aut facere repellat provident occaecati excepturi optio reprehenderit",
        "body": "quia et suscipit\nsuscipit recusandae consequuntur expedita et cum\nreprehenderit molestiae ut ut quas totam\nnostrum rerum est autem sunt rem eveniet architecto"
    }
    assert len(test.data) > 0
    assert test.data.equals(pl.from_dict(expect))


def test_transformer(prepare_test_config):
    db = DBManager('test', prepare_test_config)
    data = pl.DataFrame({
        "name": ["Alice", "Bob", "Paul"],
        "age": [20.0, 30.0, 40.0]
    })
    db.insert(data, 'users')
    data = pl.DataFrame({
        "name": ["Alice", "Bob", "Paul"],
        "height": [160, 170, 180]
    })
    db.insert(data, 'health')

    class Users(DBExtractor):
        db_name = 'test'
        query = 'SELECT * FROM users'
        _db = db
    
    class Health(DBExtractor):
        db_name = 'test'
        query = 'SELECT * FROM health'
        _db = db
    
    class TestTransformer(Transformer):
        extractor = [Users, Health]

        def process_data(self):
            users = self.data['Users']
            health = self.data['Health']
            return users.join(health, on='name')
    
    transformer = TestTransformer()
    expect = pl.DataFrame({
        "name": ["Alice", "Bob", "Paul"],
        "age": [20.0, 30.0, 40.0],
        "height": [160, 170, 180]
    })
    assert transformer.result.equals(expect)


def test_loader():
    path = Path.cwd() / 'pipeline' / 'test'
    path.mkdir(parents=True, exist_ok=True)
    
    class TestLoader(Loader):
        transformer = Transformer
        _module_setter_for_test = 'pipeline.test'
        
        def _additional_init_(self):
            pass
        
        def _set_path(self):
            pass
        
        def _load_data(self):
            pass
    
    loader = TestLoader()
    assert loader.pipe_name == 'test'
    assert loader.collection == 'test_loader'
    assert loader.location == Path('test/test_loader')


def test_dbloader(prepare_test_config):
    db = DBManager('test', prepare_test_config)
    data = pl.DataFrame({
        "name": ["Alice", "Bob", "Paul"],
        "age": [20.0, 30.0, 40.0]
    })
    db.insert(data, 'users')
    data = pl.DataFrame({
        "name": ["Alice", "Bob", "Paul"],
        "height": [160, 170, 180]
    })
    db.insert(data, 'health')

    class Users(DBExtractor):
        db_name = 'test'
        query = 'SELECT * FROM users'
        _db = db
    
    class Health(DBExtractor):
        db_name = 'test'
        query = 'SELECT * FROM health'
        _db = db
    
    class TestTransformer(Transformer):
        extractor = [Users, Health]

        def process_data(self):
            users = self.data['Users']
            health = self.data['Health']
            return users.join(health, on='name')
    
    class TestLoader(DBLoader):
        db_name = 'test'
        table_name = 'load_result'
        transformer = TestTransformer
        _db = db
        _module_setter_for_test = 'pipeline.test'
    
    TestLoader()
    data = db.select('SELECT * FROM load_result')
    expect = pl.DataFrame({
        "name": ["Alice", "Bob", "Paul"],
        "age": [20.0, 30.0, 40.0],
        "height": [160, 170, 180]
    })
    assert data.equals(expect)


@pytest.mark.skip(reason="Not yet inplemented.")
def test_fsloader():
    pass


from datetime import datetime
from dateutil.relativedelta import relativedelta
from typing import List


class PeriodGenerator:
    """
    A class for generating dates within a specific period. Returns a list of dates.

    Parameters
    ---
    - start : The start date (string)
    - end : The end date (string)
    - format : The input date format

    Example
    ---
    >>> pgen = PeriodGenerator('2020-01-01', '2020-03-01', '%Y-%m-%d')
    >>> pgen.date_list
    ['2020-01-01', '2020-02-01', '2020-03-01']
    """
    UNIT_FORMAT_MAP = {
        '%d': 'day',
        '%m': 'month',
        '%Y': 'year',
    }

    def __init__(self, start: str, end: str, format: str) -> None:
        self.format = format
        self.unit = self._parse_unit()
        self.start = self._parse_date(start)
        self.end = self._parse_date(end)

        self._validate_input()
        self.date_list = self._generate_date_list()

    def _parse_date(self, date_str: str) -> datetime:
        try:
            return datetime.strptime(date_str, self.format)
        except ValueError:
            raise ValueError(f"Invalid date '{date_str}' for format '{self.format}'.")

    def _parse_unit(self) -> str:
        for key, unit in self.UNIT_FORMAT_MAP.items():
            if key in self.format:
                return unit
        raise ValueError(f"Invalid format '{self.format}'. Supported keys are {list(self.UNIT_FORMAT_MAP.keys())}.")

    def _validate_input(self) -> None:
        if self.start > self.end:
            raise ValueError("Start date must be before or equal to end date.")

    def _generate_date_list(self) -> List[str]:
        date_list = []
        current_date = self.start

        while current_date <= self.end:
            date_list.append(current_date.strftime(self.format))
            current_date += self._get_relativedelta()

        return date_list

    def _get_relativedelta(self) -> relativedelta:
        unit_to_delta = {
            'day': relativedelta(days=1),
            'month': relativedelta(months=1),
            'year': relativedelta(years=1),
        }
        return unit_to_delta[self.unit]


from datetime import datetime
import logging
import logging.config as config
from typing import Dict, Any

from boomba.core.config import Conf
from boomba.core.constants import LOG_DIR, LOG_CONFIG
from boomba.exception.exc import LogDirectoryError


_system_log_directory_name = 'system'


class _Logger:
    _logger: logging.Logger
    log_conf: Dict[str, Any] = LOG_CONFIG
    
    def __init__(self) -> None:
        self._directory = _system_log_directory_name
        self._logger = self._configure_logger()

    def _set_file_path(self) -> None:
        file_name = f"{datetime.now().strftime('%Y%m%d')}"
        file_path = LOG_DIR / self._directory / file_name
            
        if not file_path.parent.exists():
            file_path.parent.mkdir(parents=True, exist_ok=True)
        self.log_conf['handlers']['file']['filename'] = str(file_path)
    
    def _remove_console_handler(self) -> None:
        self.log_conf['root']['handlers'].remove('console')
        
    def _configure_logger(self) -> logging.Logger:
        self._set_file_path()
        if not Conf.debug_mode:
            self._remove_console_handler()
        config.dictConfig(self.log_conf)
        return logging.getLogger()
        
    @property
    def logger(self) -> logging.Logger:
        return self._logger

    def debug(self, msg: str) -> None:
        self._logger.debug(msg)
    
    def info(self, msg: str) -> None:
        self._logger.info(msg)
        
    def warning(self, msg: str) -> None:
        self._logger.warning(msg)
    
    def error(self, msg: str) -> None:
        self._logger.error(msg)
    
    def critical(self, msg: str) -> None:
        self._logger.critical(msg)


class Logger(_Logger):

    def __init__(self, directory: str):
        if directory == 'system':
            raise LogDirectoryError(self)
        self._directory = directory
        self._logger = self._configure_logger()


from collections import deque
from dataclasses import dataclass, field
from datetime import datetime, date, timedelta, time
from enum import Enum
from pathlib import Path
import time as Time
from typing import List, Optional, Deque, Dict, Tuple

from boomba.core.config import Config, Conf
from boomba.core.constants import LOOP_INTERVAL, PIPELINE_DIR
from boomba.core.etl import Loader
from boomba.core.log import _Logger
from boomba.core.metadata import MetaDataManager
from boomba.util.parse import parse_date


__all__ = ['Job', 'register']

_metadata = MetaDataManager()


class WeekDay(Enum):
    mon = 0
    tue = 1
    wed = 2
    thu = 3
    fri = 4
    sat = 5
    sun = 6


class CronParser:
    """
    A class to parse cron expressions into individual components.

    This class takes a cron expression and splits it into its components 
    (minute, hour, day, month, and weekday). It validates the input values 
    against the allowed ranges and converts the expression into a structured 
    dictionary of possible values for each time unit.

    Attributes
    ----------
    RANGES : dict
        A dictionary defining the allowed range of values for each time unit:
        - `minute`: 0-59
        - `hour`: 0-23
        - `day`: 1-31
        - `month`: 1-12
        - `weekday`: 0-6 (Monday = 0, Sunday = 6)
    
    Methods
    -------
    parse(expr: str) -> dict:
        Parses a cron expression into a dictionary of valid values for each time unit.

    _parse_field(field: str, unit: str) -> List[int]:
        Parses a single field of a cron expression and returns a list of valid values 
        for the specified time unit.
    """
    RANGES = {
        "minute": (0, 59),
        "hour": (0, 23),
        "day": (1, 31),
        "month": (1, 12),
        "weekday": (0, 6)
    }
    
    def _split_field(self, expr: str) -> List[str]:
        fields = expr.split()
        
        if len(fields) != 5:
            raise ValueError(
                "cron expression must have 5 of fields."
            )
        return fields
    
    def _parse_field(self, field: str, unit: str) -> List[int]:
        min_v, max_v = self.RANGES[unit]
        values = set()
        
        for part in field.split(","):
            if "-" in part:
                start, end = map(int, part.split("-"))
                values.update(range(start, end + 1))
            elif "/" in part:
                base, step = part.split("/")
                step = int(step)
                
                base_range = self._parse_field(base, unit)
                
                for v in base_range:
                    while v <= max_v:
                        values.add(v)
                        v += step
            elif part == "*":
                values.update(range(min_v, max_v + 1))
            else:
                values.add(int(part))
        
        if min(values) < min_v or max(values) > max_v:
            raise KeyError(
                f"Allowed values for '{unit}' "
                f"are between {min_v} and {max_v}"
            )
        
        return list(values)
    
    def parse(self, expr: str) -> Dict[str, List[int]]:
        fields = self._split_field(expr)
            
        return {
            unit: self._parse_field(field, unit)
            for field, unit in zip(fields, self.RANGES.keys())
        }


class Schedule:
    '''
    A class to manage scheduling.

    This class generates the next scheduled time and a list of schedules 
    for the current day. The schedule representation is based on cron expressions. 
    You can either use the cron expression itself or specify individual values, 
    but using both simultaneously will result in an error.

    Since schedule building can occur multiple times on the same day 
    (e.g., every 10 minutes), if the current day matches the schedule, 
    it may generate multiple schedules for the day. This functionality 
    is implemented in the `build()` method.

    The `parser` attribute is intended for dependency injection 
    and does not need to be modified directly.
    
    Args
    ----
    The class supports both cron-specific expressions and custom values outside the default range. 
    Cron expressions allow the following special syntax:
    - `*` (all values)
    - `/` (increments)
    - `-` (ranges)
    - `,` (multiple values)

    - `minute`: Defaults to `0`, range: `0-59`.
    - `hour`: Defaults to `0`, range: `0-23`.
    - `day`: Defaults to `*`, range: `1-31`.
    - `month`: Defaults to `*`, range: `1-12`.
    - `weekday`: Defaults to `*`, range: `0-6` (0 = Monday, 6 = Sunday).
    
    Example
    -------
    >>> schedule = Schedule(expr='0/10 1-5 3 * *')
    ... schedule = Schedule(minute='0/10', hour='1-5', day=3)
    '''
    def __init__(
        self,
        minute: Optional[int|str]=None,
        hour: Optional[int|str]=None,
        day: Optional[int|str]=None,
        month: Optional[int|str]=None,
        weekday: str=None,
        expr: str=None,
        parser: CronParser=None
    ) -> None:
        self._parser = parser or CronParser()
        
        if expr is not None and any(
            param is not None 
            for param in (minute, hour, day, month, weekday)
        ):
            raise ValueError(
                "When 'expr' is set, 'minute', 'hour', 'day', "
                "'month', and 'weekday' must not be set."
            )
        
        if expr is not None:
            fields = self._parser._split_field(expr)
            self._minute = fields[0]
            self._hour = fields[1]
            self._day = fields[2]
            self._month = fields[3]
            self._weekday = fields[4]
            self._expr = expr
        else:
            self._minute = minute or 0
            self._hour = hour or 0
            self._day = day or '*'
            self._month = month or '*'
            self._weekday = self._weekday_to_int(weekday)
            self._expr = expr
    
    def __repr__(self) -> str:
        return (
            "[SchedulerBuilder]\n"
            f"- module : {self.__module__}\n"
            f"- expr : {self.expr}\n"
            f"- weeday : {self._weekday}\n"
            f"- month : {self._month}\n"
            f"- day : {self._day}\n"
            f"- hour : {self._hour}\n"
            f"- minute : {self._minute}\n"
        )
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Schedule):
            return False
        return self.expr == other.expr
    
    @property
    def year(self) -> int:
        return datetime.today().year
    
    @property
    def expr(self) -> str:
        if self._expr is None:
            return (
                f"{self._minute} "
                f"{self._hour} "
                f"{self._day} "
                f"{self._month} "
                f"{self._weekday}"
            )
        
        return self._expr
    
    def _weekday_to_int(self, weekday: str) -> int:
        if weekday is None:
            return '*'
        
        if '/' in weekday:
            raise ValueError("The weekday cannot contain the '/' character.")
        
        for k, v in WeekDay.__members__.items():
            weekday = weekday.replace(k, str(v.value))
        
        return weekday
    
    def _is_weekday_match(self, date: datetime) -> Optional[bool]:
        valid_days = self._parser._parse_field(self._weekday, 'weekday')
        return date.weekday() in valid_days
    
    def next(self, base: datetime=None) -> datetime:
        """
        Calculates the next scheduled time based on the provided base time.

        Args
        ----
            base (datetime): The base time from which the next schedule is calculated.
                             Defaults to the current datetime if not provided.

        Returns
        -------
            datetime: The next scheduled datetime.
        """
        parsing_data = self._parser.parse(self.expr)
        
        if base is None:
            base = datetime.now()
        
        keys = {
            'minute': 0,
            'hour': 1,
            'day': 2,
            'month': 3,
            'year': 4
        }
        to_update = set()
        values = {'year': base.year}
        
        for key, id in keys.items():
            if id == (len(keys) - 1):
                if id in to_update:
                    values[key] += 1
                # year doesn't exist in parsing_data
                break
            
            parsing_data[key].sort() # ensure time sequence
            base_value = getattr(base, key)
            
            # The smallest unit is a minute.
            # Only update the value when equal to or greater than the base value, skipping other steps.
            if id == 0 or id in to_update:
                valid_value = min(
                    [v for v in parsing_data[key]
                    if v > base_value],
                    default=None
                )
            else:
                valid_value = min(
                    [v for v in parsing_data[key]
                    if v >= base_value],
                    default=None
                )
            
            if valid_value is None:
                valid_value = min(parsing_data[key])
                to_update.add(id + 1)
            
            values[key] = valid_value
        
        # when upper changed, lower reset
        if max(to_update, default=None) is not None:
            to_min = [k for k, v in keys.items() if v < max(to_update)]
        
            for k in values.keys():
                if k in to_min:
                    values[k] = min(parsing_data[k])
        
        # result
        result = base.replace(
            year=values['year'],
            month=values['month'],
            day=values['day'],
            hour=values['hour'],
            minute=values['minute'],
            second=0,
            microsecond=0
        )
        
        # check weekday
        if self._weekday != '*':
            if not self._is_weekday_match(result):
                result = self.next(result)
        
        return result
    
    def build(self) -> Optional[List[datetime]]:
        """
        Builds a list of schedules for the current day, starting from midnight.

        Returns:
            Optional[List[datetime]: A list of schedules for the current day.
            If the current day does not match the schedule, returns `None`.
        """
        today = datetime.today()
        tommorow = datetime.combine(today, time.min) + timedelta(days=1)
        
        if self._weekday != '*':
            if not self._is_weekday_match(today):
                return None
        
        next_ = self.next()
        schedule: Optional[List[datetime]] = []
        
        while True:
            if next_ >= tommorow or next_ is None:
                break
            
            schedule.append(next_)
            next_ = self.next(next_)
        
        return schedule


def _register_jobs() -> None:
    '''Force execute the schedule.py file of all pipelines to register the jobs.'''
    try:
        path = Path(PIPELINE_DIR)
        for dir in path.iterdir():
            if dir.is_dir() and dir.name != '__pycache__':
                file = dir / 'schedule.py'
                
                with open(file, 'r') as f:
                    schedule = f.read()
                exec(schedule, {}, {})
    
    except Exception as e:
        raise FileNotFoundError(str(e))


@dataclass(eq=False)
class ETLJob:
    loader: Loader
    name: str
    schedule: Schedule
    start: datetime = field(default_factory=datetime.now)
    end: datetime = field(default=None)
    completed_jobs: List[datetime] = field(default_factory=list)
    next_job: datetime = field(default=None)
    today_jobs: List[datetime] = field(default_factory=list)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ETLJob):
            return False
        return (
            self.loader == other.loader and
            self.schedule == other.schedule
        )
        
        
class JobMeta(type):
    _instances = {}
    
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]
        

class Job(metaclass=JobMeta):
    _logger: _Logger

    def __init__(
        self,
        conf: Config=Conf,
        schedule: Schedule=None,
        metadata: MetaDataManager=_metadata
    ) -> None:
        self._conf = conf
        self._schedule = schedule or Schedule
        self.job_list: List[ETLJob] = []
        self._metadata = metadata
        self._logger = _Logger()
    
    def __str__(self) -> str:
        return f"Job with {len(self.job_list)} jobs"
    
    def __repr__(self) -> str:
        return (
            f"- jobs : {len(self.job_list)}\n"
            f"- job_list : {[j.name for j in self.job_list]}\n"
        )
    
    @property
    def now(self) -> datetime:
        return datetime.now()
    
    @property
    def today(self) -> datetime:
        return datetime.combine(
            date.today(),
            datetime.min.time()
        )
    
    def _filter_active_jobs(self, job_list: List[ETLJob]) -> List[ETLJob]:
        '''Filter jobs that are active based on their start and end dates.'''
        new_list: List[ETLJob] = []
        
        for job in job_list:
            start_ = job.start is not None
            end_ = job.end is not None
            
            if start_ and not end_:
                if job.start > self.now:
                    continue
            
            if not start_ and end_:
                if job.end < self.now:
                    continue
            
            if all((start_, end_)):
                if job.start > self.now and job.end < self.now:
                    continue
            
            new_list.append(job)
        
        return new_list
    
    def _initiate_schedule(self, job_list: List[ETLJob]) -> List[ETLJob]:
        job_list = self._filter_active_jobs(job_list)
        self._metadata.init_schdule(self.now)
        
        new_list: List[ETLJob] = []
        for job in job_list:
            schedule = job.schedule.build()
            
            if schedule:
                job.today_jobs = schedule
                job.next_job = job.schedule.next()
                new_list.append(job)
                
        return new_list
    
    def _initiate_queue(
        self,
        job_list: List[ETLJob]
    ) -> Deque[Tuple[Loader, datetime, str]]:
        '''
        Initializes a job queue based on the given job list.
    
        The queue is populated with tuples consisting of the loader, scheduled time, 
        and job name, sorted by scheduled time.
        
        Returns
        -------
        Deque[Tuple[Loader, datetime]]
        A queue of tuples, each containing a loader, scheduled time, and job name, 
        sorted by scheduled time.
        '''
        tmp_list: List[Tuple[Loader, datetime, str]] = []
        for job in job_list:
            for t in job.today_jobs:
                tmp_list.append((job.loader, t, job.name))
                
                self._metadata.save_schedule(
                    scheduled_at=t,
                    job_name=job.name
                )
                
        tmp_list.sort(key=lambda x: x[1]) # to ensure time sequence
        q = deque(tmp_list)
        self._metadata.save_queue(q)
        return q
    
    def _print_status(self) -> None:
        n = len(self.job_list)
        job_names = [job.name for job in self.job_list]
        print(
            f"Registered {n} jobs in the system. "
            f"The job List={', '.join(job_names)}"
        )
    
    def register(
        self,
        loader: Loader,
        name: str,
        minute: int=None,
        hour: int=None,
        day: int=None,
        month: int=None,
        weekday: str=None,
        expr: str=None,
        start: str=None,
        end: str=None,
    ) -> None:
        schedule = self._schedule(
            minute=minute,
            hour=hour,
            day=day,
            month=month,
            weekday=weekday,
            expr=expr
        )
        format = self._conf.job_date_format
        start_ = parse_date(start, format)
        end_ = parse_date(end, format)
        
        job = ETLJob(
            loader=loader,
            name=name,
            schedule = schedule,
            start=start_,
            end=end_
        )
        
        for j in self.job_list:
            if name == j.name:
                raise ValueError(f"Duplicate job name. {name}")
        
        if job not in self.job_list:
            self.job_list.append(job)
            self._metadata.save_job(
                job_name=job.name,
                registered_at=datetime.now(),
                start_date=job.start,
                end_date=job.end,
                cron_expression=expr
            )
        else:
            print(f"[Warning] '{name}' is already in the list.")
    
    def run(self) -> None:
        init_date = None
        _register_jobs()
        self._print_status()

        while True:
            Time.sleep(LOOP_INTERVAL)
            
            # initiate job every day(or when run first)
            if init_date is None or init_date.day != self.now.day:
                init_date = self.now
                job_list = self._initiate_schedule(self.job_list)
                job_queue = self._initiate_queue(job_list)
            
            if not job_queue:
                print(f'all jobs have been done... {self.now}')
                continue
            
            if job_queue[0][1] <= self.now:
                job = job_queue.popleft()
                
                try:
                    inst: Loader = job[0]() # run Loader subclass
                    
                    for i, j in enumerate(job_list):
                        if j.name == job[2]:
                            job_list[i].completed_jobs.append(job[1])
                            job_list[i].next_job = j.schedule.next()
                            print(f'job completed... {j.name}')
                    
                    # save the result of the job execution into the default database
                    self._metadata.save_schedule(
                        scheduled_at=job[1],
                        job_name=job[2],
                        executed_at=datetime.now(),
                        is_completed=True
                    )
                    self._metadata.save_dir(
                        pipe_name=inst.pipe_name,
                        collection=inst.collection,
                        job_name=job[2]
                    )
                    self._metadata.save_path(
                        file_name=inst.file_name,
                        pipe_name=inst.pipe_name,
                        collection=inst.collection,
                        scheduled_at=job[1]
                    )
                    self._metadata.save_queue(job_queue) # renew the table of queue
                    
                except Exception as e:
                    self._logger.error(e)
                    print(f"[ERROR] {e}")
                

def register(
    loader: Loader,
    name: str,
    minute: int=None,
    hour: int=None,
    day: int=None,
    month: int=None,
    weekday: str=None,
    expr: str=None,
    start: str=None,
    end: str=None,
) -> None:
    '''API for user convenience when registering a job'''
    job = Job()
    job.register(
        loader=loader,
        name=name,
        minute=minute,
        hour=hour,
        day=day,
        month=month,
        weekday=weekday,
        expr=expr,
        start=start,
        end=end
    )


from typing import List, Union

import polars as pl
from sqlalchemy import create_engine, text
from sqlalchemy.engine import Row, Engine
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import SQLAlchemyError

from boomba.core.config import Conf, Config
from boomba.exception.exc import DBConnectionError


__all__ = ['DBManager']


class Connector:
    """
    A database connector that creates and manages SQLAlchemy engines.

    This class uses DBConfig to load and validate configuration settings, 
    and then generates connection strings and SQLAlchemy engines for database interaction.

    Attributes:
        __conf (DBConfig): The configuration object for the database connection.

    Args:
        section (str): The name of the configuration section to load.
    """

    def __init__(self, db_name: str, conf: Config=None, **kwagrs) -> None:
        self._db_name = db_name
        self._conf = conf or Conf
        self.engine = self._create_engine(**kwagrs)
    
    def _get_conn_str(self) -> str:
        try:
            fields = self._conf.database[self._db_name]
            return URL.create(**fields)
        except:
            raise DBConnectionError(self)

    def _create_engine(self, **kwagrs) -> Engine:
        connection_string = self._get_conn_str()
        try:
            engine = create_engine(connection_string, **kwagrs)
            return engine
        except RuntimeError as e:
            raise DBConnectionError(self, e)


class DBManager(Connector):
    """
    A database manager that provides methods for executing SELECT, UPDATE, DELETE, and INSERT queries.
    This class also supports transaction handling.

    Parameter
    ---
    section (str): The name of the configuration section to load.

    Example:
        ```python
        # Example usage
        db_manager = DBManager("my_database_section")
        query = "SELECT name, age FROM some_table"
        db_manager.select(query)
        ```
    Attributes
    ---
    - engine (Engine): The SQLAlchemy engine for database connection.
    """

    def __init__(self, db_name: str, conf: Config=None):
        super().__init__(db_name, conf)
        
    def execute_query(self, query: str, params: dict = None) -> None:
        """
        Executes a raw SQL query on the database.
        
        Parameters
        ---
        - query (str): The SQL query to be executed.
        - params (dict, optional): A dictionary of parameters to bind to the query. Defaults to None.

        Example
        ---
        ```python
        db_manager = DBManager("my_database_section")
        query = "UPDATE users SET age = :age WHERE name = :name"
        params = {"name": "Alice", "age": 30}
        db_manager.execute(query, params)
        ```
        
        Notes
        ---
        - This method does not return any data; it simply executes the provided SQL query.
        - If the query is a SELECT, it will not return results.
        - If the query involves a modification (INSERT, UPDATE, DELETE), it will be committed immediately.
        """
        with self.engine.begin() as connection:
            connection.execute(text(query), params or {})

    def select(self, query: str, params: dict = None, to_df: bool=True) -> Union[pl.DataFrame, List[Row]]:
        """
        Executes a SELECT query and returns the results, either as a list of rows or a Polars DataFrame.

        Parameters
        ----------
        - query (str): The SQL query to execute.
        - params (dict, optional): A dictionary of parameters to bind to the query. Defaults to None.
        - to_df (bool, optional): If True, the result will be returned as a Polars DataFrame. If False, a list of rows is returned. Defaults to True.
            
        Example
        -------
        ```python
        db_manager = DatabaseManager(connector)
        query = "SELECT * FROM users WHERE age > :age"
        params = {"age": 30}
        results = db_manager.select(query, params, to_df=True)
        print(results)
        ```

        Returns
        -------
        - If `to_df` is True, returns a Polars DataFrame containing the query result.
        - If `to_df` is False, returns a list of rows (each row is a SQLAlchemy Row object).

        Raises
        ------
        - RuntimeError: If there is an error during the SQL query execution.
        """
        with self.engine.connect() as connection:
            result = connection.execute(text(query), params or {})
            
            if to_df:
                return self._to_DataFrame(result)
            
            return result.fetchall()

    def insert(
            self,
            df: pl.DataFrame,
            table_name: str,
            if_exists: str = "append",
            chunksize: int = 100000
        ) -> None:
        """
        Inserts a Polars DataFrame into a database table using Polars' write_database method.

        Parameters
        ----------
        df : pl.DataFrame
            The Polars DataFrame to insert into the database.
        table_name : str
            The name of the target table in the database.
        if_exists : str, optional
            Behavior when the table already exists. Options are:
            - 'fail': raises an error (default)
            - 'replace': drops the table and recreates it
            - 'append': appends the data to the existing table
        chunksize : int, optional
            The number of rows to insert in each batch. Default is 100000.

        Example
        -------
        ```python
        df = pl.DataFrame({"id": [1, 2], "name": ["Alice", "Bob"]})
        db_manager = DBManager("mydb")
        db_manager.insert(df, "users", if_exists="append")
        ```

        Raises
        ------
        RuntimeError
            If there is an error while inserting the data into the database.
        """
        try:
            if not isinstance(df, pl.DataFrame):
                raise TypeError(
                    f"Expected a Polars DataFrame, but got '{type(df).__name__}' instead."
                )
            
            df.write_database(
                table_name=table_name,
                connection=self.engine,
                if_table_exists=if_exists,
                engine="sqlalchemy",
                engine_options={"chunksize": chunksize}
            )
        except Exception as e:
            raise RuntimeError(f"Failed to insert DataFrame into {table_name}: {e}")

    def transaction(self, queries: list):
        """
        Executes a list of queries within a transaction.

        Parameters
        ---
        - queries (list): A list of dictionaries containing 'query' and optional 'params'.
        
        Example
        ---
        ```python
            # Example usage of the transaction method
            queries = [
                {'query': 'INSERT INTO users (name, age) VALUES (:name, :age)', 'params': {'name': 'Alice', 'age': 30}},
                {'query': 'UPDATE users SET age = :age WHERE name = :name', 'params': {'name': 'Alice', 'age': 31}},
            ]
            try:
                db_manager = DatabaseManager(connector)
                db_manager.transaction(queries)
            except RuntimeError as e:
                print(f"Error: {e}")
        ```
        
        Raises:
            SQLAlchemyError: If an error occurs during the transaction.
        """
        with self.engine.begin() as connection:
            try:
                for q in queries:
                    connection.execute(text(q['query']), q.get('params', {}))
            except SQLAlchemyError as e:
                raise RuntimeError(f"Transaction failed: {e}")
            
    def _to_DataFrame(self, result: List[Row]) -> pl.DataFrame:
        """
        Converts a list of SQL query results into a Polars DataFrame.

        This method processes the result of a SQL query, extracts the column names, and converts the result
        into a Polars DataFrame.

        Parameters
        ----------
        - result (list): The query result as a list of rows (usually returned by SQLAlchemy).

        Example
        -------
        ```python
        query = "SELECT id, name FROM users"
        result = db_manager.select(query)
        df = db_manager._to_DataFrame(result)
        print(df)
        ```

        Returns
        -------
        - pl.DataFrame: A Polars DataFrame containing the query result.

        Raises
        ------
        - ValueError: If no data is returned from the query.
        - Exception: If there is an error while creating the Polars DataFrame.
        """
        if not result:
            raise ValueError("No data returned from the query.")
        
        try:
            return pl.DataFrame(result)
        except Exception as e:
            raise ValueError(f"Failed to create Polars DataFrame: {e}")


from typing import Dict, Any
from pathlib import Path
from boomba.core.constants import CONF_PATH, MOCK_CONFIG


class Config:
    is_mock: bool
    conf_path: Path
    
    def __init__(self, conf_path: Path=None) -> None:
        self.is_mock = False
        self.conf_path = conf_path or Path(CONF_PATH)
        self._vars = self._load()
        self._set_attr()
    
    def __repr__(self) -> str:
        return (
            f"Config(var={self._vars}, is_mock={self.is_mock})"
        )
    
    def _load(self) -> Dict[str, Any]:
        if not self.conf_path.exists():
            return self._temp_config()
        
        with open(str(self.conf_path)) as f:
            conf = f.read()
        
        vars = {}
        exec(conf, {}, vars)
        return vars
    
    def _temp_config(self) -> Dict[str, Any]:
        self.is_mock = True
        return MOCK_CONFIG
    
    def _set_attr(self) -> None:
        for k, v in self._vars.items():
            if not k.startswith('__'):
                setattr(self, k.lower(), v)


Conf = Config()


import sys
import unittest
from unittest.mock import patch

from argparse import ArgumentParser, Namespace
from io import StringIO

from boomba.cli.base import BaseCommand
from boomba.cli.description import ERROR


class TestBaseCommand(unittest.TestCase):
    
    def setUp(self):
        sys.argv = ['test', '--name', 'test_name']
        class TempCommand(BaseCommand):
            _description = 'test description'
            _epilog = 'test epilog'

            def add_all(self):
                self._parser.add_argument(
                    "--name",
                    required=True
                )
        
        self.inst = TempCommand()
    
    def test_create_parser(self):
        self.assertIsInstance(self.inst._parser, ArgumentParser)
        self.assertEqual(self.inst._parser.description, 'test description')
        self.assertEqual(self.inst._parser.epilog, 'test epilog')
    
    def test_create_subparser(self):
        self.assertIsNotNone(self.inst._subparser)
    
    def test_add_all_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            class InCompleteCommand(BaseCommand):
                _description = 'test_incoplete_description'
                _epilog = 'test_incomplete_epilog'
                pass
            InCompleteCommand()
    
    @patch('argparse.ArgumentParser.parse_args', return_value=Namespace(name='test_name'))
    def test_parse(self, mock):
        args = self.inst.parse()
        self.assertEqual(args.name, 'test_name')
    
    @patch('argparse.ArgumentParser.parse_args', return_value=Namespace(command='test_commnad'))
    def test_get_commnad(self, mock):
        self.inst._args = mock.return_value
        self.assertEqual(self.inst.get_command(), 'test_commnad')
    
    @patch('argparse.ArgumentParser.parse_args', return_value=Namespace(test='test_option'))
    def test_get_option_vaild_case(self, mock):
        self.inst._args = mock.return_value
        self.assertEqual(self.inst.get_option('--test'), 'test_option')
    
    @patch('argparse.ArgumentParser.parse_args', return_value=Namespace(test=None))
    @patch('sys.stdout', new_callable=StringIO)
    def test_get_option_invalid_case(self, mock_out, mock_args):
        self.inst._args = mock_args.return_value
        self.assertIsNone(self.inst.get_option('--test'))
        self.assertIn(ERROR['empty_args'], mock_out.getvalue())


from typing import Dict
from boomba.core.constants import PYTHON_VERSION, BOOMBA_VERSION


HEADER: Dict[str, str] = {
    'description': (
        "Boomba is a lightweight and efficient ETL pipeline framework.\n\n"
        "Features:\n"
        "- Schedule and automate ETL jobs\n"
        "- Support for data extraction, transformation, and loading tasks\n"
        "- Build efficient data pipelines with ease\n\n"
        "[Example Usage]:\n"
        " $ boomba startproject\n"
        " $ boomba run\n"
        " $ boomba createpipe <pipe_name>\n"
    ),
    
    'epilog': (
        "[Detailed Job Information]\n"
        "If you want to check more detailed job statuses,\n"
        "please connect to the database specified in the configuration file.\n"
        "In the database, you can view the execution time and results of jobs,\n"
        "the paths of generated files, the status of the current job queue,\n"
        "and the list of jobs created today. \n\n"
        "For more information,\n"
        " - github : https://github.com/Baboomba/boomba\n"
        " - email  : bach0918@gmail.com"
    ),
}

HELP: Dict[str, str] = {
    'startproject': 'Create boomba project(no additional arguments)',
    
    'initdb': 'Initiate database for boomba systme(not for extraction)',
    
    'run': 'Execute boomba application(no additional arguments)',
    
    'createpipe': (
        "Create a new pipeline (requires a string argument,\n"
        "the additional argument is set as the pipeline name)"
    ),
    
    'name': 'Pipeline name to create',
    
    'test': 'Execute a specific pipeline directly (for development purposes)'
}

MESSAGE: Dict[str, str] = {
    'initial': (
        "=============================================\n"
        f"Boomba v{BOOMBA_VERSION} = Startup\n"
        "=============================================\n"
        "Initializing Boomba framework...\n"
        f"Python version: {PYTHON_VERSION}\n"
    ),
    
    'startproject': 'Project creation is complete.',

    'initdb': 'Default database has been initialized.',

    'start_test': "Start testing the job '{}'...",
    'end_test': "Completed testing."
}

CHECK: Dict[str, str] = {
    'data_dir': "Default data directory: {}",
    
    'database': 'Base database has successfully connected to IP adress : {}',
    
    'table_exist': 'tables already exist. [Table] {}',
    
    'start': 'Please start a project first.'
}

ERROR: Dict[str, str] = {
    'invalid_args': "Invalid argument. Please, read help(--help)",
    'empty_args': (
        "No arguments were provided. Please refer to the help.\n"
        " $ boomba --help"
    ),
    'invalid_job_name': "Invalid job name. Ther is no such job '{}'."
}


from string import Template
from typing import Dict
from pathlib import Path

from boomba.core.constants import (
    DATA_DIR,
    LOG_DIR,
    CONF_DIR,
    PIPELINE_DIR,
    TPL_DIR
)


class Generator:
    directories: Dict[str, Path] = {
        'data': DATA_DIR,
        'log': LOG_DIR,
        'conf': CONF_DIR
    }
    
    conf_path: Dict[str, Path] = {
        'settings': TPL_DIR / 'settings.py-tpl'
    }
    
    pipe_dir: Path = TPL_DIR / 'pipeline'
    pipe_path: Dict[str, str] = {
        'extraction': pipe_dir / 'extraction.py-tpl',
        'load': pipe_dir / 'load.py-tpl',
        'schedule': pipe_dir / 'schedule.py-tpl',
        'schema': pipe_dir / 'schema.py-tpl',
        'transformation': pipe_dir / 'transformation.py-tpl',
    }
    
    def convert_template(
        self,
        file_path: str,
        output_path: str,
        context: Dict[str, str]=None
    ) -> None:
        with open(file_path, 'r') as tpl_file:
            tpl_file = tpl_file.read()
        
        if context is None:
            context = {}
            
        template = Template(tpl_file)
        output = template.substitute(context)

        with open(output_path, "w") as file:
            file.write(output)
            
    def init_directories(self) -> None:
        for _, dir in self.directories.items():
            Path(dir).mkdir(parents=True, exist_ok=False)
        
        for k, v in self.conf_path.items():
            self.convert_template(
                file_path=v,
                output_path=CONF_DIR / f"{k}.py"
            )
        
    def create_pipeline(self, pipe_name: str) -> None:
        dir = PIPELINE_DIR / pipe_name
        try:
            dir.mkdir(parents=True, exist_ok=False)
        except Exception as e:
            raise FileExistsError(f"'{pipe_name}' already exists.") from e
        for k, v in self.pipe_path.items():
            self.convert_template(
                file_path=v,
                output_path=dir / f"{k}.py"
            )


from pathlib import Path
from typing import Dict, Any

PYTHON_VERSION = '3.13.2'
BOOMBA_VERSION = '0.1.0b5'

# The pakage directory
BOOMBA_DIR: Path = Path(__file__).resolve().parent.parent

# template directory
TPL_DIR = BOOMBA_DIR / 'template'


# The project directory of users
BASE_DIR: Path = Path.cwd()

# The directory and path of 'setting.py'
CONF_DIR = BASE_DIR / 'config'
CONF_PATH = BASE_DIR / CONF_DIR / 'settings.py'

# The directory of pipeline
PIPELINE_DIR  = BASE_DIR / 'pipeline'

# The default directory of data. it can be changed by users.
DATA_DIR = BASE_DIR / 'data'

# The directory of log
LOG_DIR = BASE_DIR / 'log'

# The time interval of run-loop. The unit is second.
LOOP_INTERVAL: int = 1

# The mode for writing the log file.
LOG_MODE: str = 'a'

# The encoding rule for writing
ENCODING: str = 'utf-8'

# The base file format for loading
BASE_DATA_FORMAT: str = 'parquet'

# The base date format for the attributes, start_date, and end_date of Loader class
DATE_FORMAT: str = 'YYYYMMDD'

# for APIExtractor
ACCEPT_HEADERS: Dict[str, str] = {
    "Accept": "application/json"
}

# for APIExtractor
CONTENT_HEADERS: Dict[str, str] = {
    "Content-Type": "application/json"
}

# To prevent errors when creating the Config() object during the first command execution
MOCK_CONFIG: Dict[str, Any] = {
    'database': {
        'mock_db': {
            'drivername': 'sqlite',
            'database': 'db',
        }
    },
    'base_db': 'mock_db',
}

LOG_CONFIG: Dict[str, Any] = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s [%(levelname)s] %(message)s'
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'standard',
            'stream': 'ext://sys.stdout'
        },
        'file': {
            'class': 'logging.FileHandler',
            'level': 'WARNING',
            'formatter': 'standard',
            'filename': ''
        }
    },
    'root': {
        'level': 'DEBUG',
        'handlers': ['console', 'file']
    }
}


import polars as pl
from boomba.core.dtypes import Dtype
from boomba.core.schema import Schema


def test_schema():
    expected_schema = [
        ('int8', pl.Int8),
        ('int16', pl.Int16),
        ('int32', pl.Int32),
        ('int64', pl.Int64),
        ('uint8', pl.UInt8),
        ('uint16', pl.UInt16),
        ('uint32', pl.UInt32),
        ('uint64', pl.UInt64),
        ('float32', pl.Float32),
        ('float64', pl.Float64),
        ('string', pl.String),
        ('boolean', pl.Boolean),
        ('date_', pl.Date),
        ('time_', pl.Time),
        ('datetime_', pl.Datetime),
        ('decimal', pl.Decimal),
    ]
    
    class TestSchema(Schema):
        int8 = Dtype.Int8()
        int16 = Dtype.Int16()
        int32 = Dtype.Int32()
        int64 = Dtype.Int64()
        uint8 = Dtype.UInt8()
        uint16 = Dtype.UInt16()
        uint32 = Dtype.UInt32()
        uint64 = Dtype.UInt64()
        float32 = Dtype.Float32()
        float64 = Dtype.Float64()
        string = Dtype.String()
        boolean = Dtype.Boolean()
        date_ = Dtype.Date()
        time_ = Dtype.Time()
        datetime_ = Dtype.Datetime()
        decimal = Dtype.Decimal(10, 2)
    
    assert list(TestSchema().schema.items()) == expected_schema


import inspect

import polars as pl

from boomba.core.dtypes import _Dtypes


class _MetaSchema(type):
    '''
    A metaclass that automatically generates a PyArrow schema from the attributes 
    of a class that are instances of _Dtypes.

    This metaclass will scan the class attributes and check if any of them 
    are instances of the _Dtypes class or its subclasses. If so, the corresponding
    PyArrow data type will be used to build a schema that is assigned to the 
    class as the schema attribute.

    Attributes:
        name (str): The name of the class being created.
        bases (tuple): A tuple of base classes of the class being created.
        dct (dict): A dictionary of class attributes.
    
    Returns:
        type: The class being created, with the added `schema` attribute.
    '''
    def __new__(cls, name, bases, dct):
        schema = []
        
        if bases:
            for k, v in dct.items():
                if k.startswith('__') and k.endswith('__'):
                    continue
                
                for base in bases:
                    if hasattr(base, k):
                        continue
                
                if inspect.isfunction(v):
                    continue
                
                if isinstance(v, _Dtypes):
                    schema.append((k, v.dtype))
                else:
                    raise TypeError(
                        f"Invalid attribute '{k}': Expected instance of "
                        f"_Dtypes, got {type(v).__name__}."
                    )
        
        dct['schema'] = pl.Schema(schema)
        return super().__new__(cls, name, bases, dct)


class Schema(metaclass=_MetaSchema):
    '''
    A base class that uses BuilderMeta as its metaclass. 

    Classes inheriting from BaseBuilder should define their attributes as instances 
    of subclasses of _Dtypes, which will automatically be used to generate the 
    PyArrow schema.

    Example:
        class CustomSchema(Schema):
            name = String()
            age = Int32()
    
    In this example, a PyArrow schema will be automatically generated for the 
    'name' and 'age' attributes.
    '''
    pass





def create_dependecty(obj: object, _type: object):
    if obj is None:
        return obj
    else:
        if type(obj) != _type:
            raise TypeError(
                f"Invalid parameter: The '{obj}' argument "
                f"must be of type '{type(obj)}'."
                f""
            )
        return obj


from pathlib import Path
import warnings

from sqlalchemy import inspect

from boomba.cli.base import BaseCommand
from boomba.cli.description import (
    HEADER,
    HELP,
    MESSAGE,
    CHECK,
    ERROR
)
from boomba.core.config import Config
from boomba.core.constants import DATA_DIR
from boomba.core.db import Connector
from boomba.core.metadata import Base
from boomba.core.schedule import Job, _register_jobs
from boomba.exception.exc import (
    NonEmptyDirectoryError,
    DatabaseConfigurationError,
    EmptyArgumentError
)
from boomba.template.template import Generator




class CommandRegistor(BaseCommand):
    _description = HEADER['description']
    _epilog = HEADER['epilog']
    
    # commands
    startproject: str = 'startproject'
    run: str = 'run'
    createpipe: str = 'createpipe'
    name: str = 'name'
    initdb: str = 'initdb'
    test: str = 'test'
    
    # valid options
    _list_default = 'all'
    
    def add_all(self) -> None:
        self.add_startproject()
        self.add_run()
        self.add_createpipe()
        self.add_initdb()
        self.add_test()
    
    def add_startproject(self) -> None:
        self._subparser.add_parser(
            self.startproject,
            help=HELP[self.startproject]
        )
    
    def add_initdb(self) -> None:
        self._subparser.add_parser(
            self.initdb,
            help=HELP[self.initdb]
        )
    
    def add_run(self) -> None:
        self._subparser.add_parser(
            self.run,
            help=HELP[self.run]
        )
    
    def add_createpipe(self) -> None:
        self._subparser.add_parser(
            self.createpipe,
            help=HELP[self.createpipe]
        ).add_argument(
            self.name,
            type=str,
            help=self.name
        )
    
    def add_test(self) -> None:
        self._subparser.add_parser(
            self.test,
            help=HELP[self.test]
        ).add_argument(
            self.name,
            type=str,
            help=self.name
        )


class CommandHandler(CommandRegistor):
    
    def __init__(self):
        super().__init__()
        self.generator = Generator()
        self.handle()
    
    def handle(self) -> None:
        command = self.get_command()
        
        if command == self.run:
            self.run_command()
        elif command == self.createpipe:
            self.createpipe_command()
        elif command == self.startproject:
            self.startproject_command()
        elif command == self.initdb:
            self.initdb_command()
        elif command == self.test:
            self.test_command()
        else:
            raise EmptyArgumentError()
    
    def startproject_command(self) -> None:
        if any(Path.cwd().iterdir()):
            raise NonEmptyDirectoryError()
        
        self.generator.init_directories()
        print(MESSAGE['startproject'])
    
    def initdb_command(self) -> None:
        conf = Config()
        if conf.is_mock:
            warnings.warn(CHECK['start'])
            return
        
        connector = Connector(conf.base_db, echo=True)
        
        ins = inspect(connector.engine)
        tables = ins.get_table_names()
        if tables:
            warnings.warn(CHECK['table_exist'].format(tables))
        
        Base.metadata.create_all(connector.engine)
        print(MESSAGE['initdb'])
    
    def run_command(self) -> None:
        print(MESSAGE['initial'])
        conf = Config()
        
        # get the load directory
        print(CHECK['data_dir'].format(DATA_DIR))
        
        # check db
        db = conf.database[conf.base_db]
        if db['drivername'] == 'sqlite':
            db_ip = 'sqlite:///' + db['database']
        else:
            db_ip = db['host']
        
        connector = Connector(conf.base_db)
        try:
            with connector.engine.connect() as conn:
                conn.close()
                print(CHECK['database'].format(db_ip))
        except:
            raise DatabaseConfigurationError()
        
        job = Job()
        job.run()
    
    def createpipe_command(self) -> None:
        self.generator.create_pipeline(self._args.name)
    
    def test_command(self) -> None:
        _register_jobs()
        job = Job()
        for j in job.job_list:
            if j.name == self._args.name:
                print(MESSAGE['start_test'].format(j.name))
                try:
                    j.loader()
                    print(MESSAGE['end_test'])
                    return
                except RuntimeError as e:
                    raise print(e)
        print(ERROR['invalid_job_name'].format(self._args.name))


from setuptools import setup, find_packages

setup(
    name="boomba",
    version="0.1.0b5",
    packages=find_packages(),
    install_requires=['polars', 'pyarrow', 'SQLAlchemy'],
    entry_points={
        'console_scripts': [
            'boomba=boomba.main:main',
        ]
    },
    author="SungEun An",
    author_email="bach0918@gmail.com",
    description='Light and Fast ETL Framework',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Baboomba/boomba',
    python_requires='>=3.13'
)


from collections.abc import Iterable
from datetime import datetime
import re
from typing import Optional


def to_snake_case(string: str) -> str:
    result = re.sub(r'([a-z])([A-Z])', r'\1_\2', string)
    return result.lower()


def to_date_format(format: str, valid_specials: Iterable=None) -> str:
    """
    Converts a given date format string (e.g., 'YYYY-MM-DD HH:mm:ss') to 
    a format recognized by Python's strftime (e.g., '%Y-%m-%d %H:%M:%S').
    
    Args
    ----
    date_format : str
        The date format string containing custom date placeholders like 
        'YYYY', 'MM', 'DD', etc., that need to be converted.
    
    Returns
    -------
    str
        The date format string converted to the Python strftime format.
    
    Raises
    ------
    ValueError
        If the provided date format contains unsupported or invalid characters.
    
    Example
    -------
    >>> to_date_format('YYYY-MM-DD HH:mm:ss')
    '%Y-%m-%d %H:%M:%S'
    """
    date_map = {
        'YYYY': '%Y',
        'MM': '%m',
        'DD': '%d',
        'HH': '%H',
        'mm': '%M',
        'ss': '%S'
    }
    
    if valid_specials is None:
        valid_specials = "-_ "
    
    check = format
    
    # Remove map key
    for key in date_map:
        if key in check:
            index = check.find(key)
            check = check[:index] + check[index + len(key):]
    
    # Remove allowed special characters
    for char in valid_specials:
        check = check.replace(char, "")
    
    # If there are any remaining characters, raise an error
    if check:
        raise ValueError(
            f"Invalid format found in the date_format "
            f"of configuration: {check}"
        )
            
    for k, v in date_map.items():
        format = format.replace(k, v)
    
    return format


def parse_date(
        date_str: str,
        format: str,
        valid_specials: Iterable=None,
        time_only: bool=False
    ) -> Optional[datetime]:
    '''
    Converts a string in YYYYMMDD HHmmss format to a datetime object.

    Args
    ----
    - date_str : The date string. For example: 20250101, 16:20:00
    - format : The date format in YYYYMMDD HHmmss style
    - valid_specials : Allowed special characters. Default is '-', '_', ' '
    - time_only : Returns only the time without the year, month, and day. Default is False
    
    Example
    -------
    >>> date = parse_date('16:20:00', 'HH:mm:ss', ':')
    >>> print(date)
    ... datetime.time(16, 20)

    See also
    --------
    to_date_format
    '''
    if date_str is not None:
        format = to_date_format(format, valid_specials)
        new_date = datetime.strptime(date_str, format)
    
        if time_only:
            return new_date.time()
        
        return new_date
    
    return None


def parse_boolean(key: str) -> bool:
    """
    Parses and validates the boolean field to determine the file system type.
    
    Args
    ----
    - key(str) : The field to check
    
    Returns
    -----
    - bool: True if the file system is remote, False if local.
    
    Raises:
    - ValueError: If the 'is_remote' field is not set to 'true' or 'false'.
    """
    valid_values = {'true', 'false'}
    key = key.lower()
    
    if key not in valid_values:
        raise ValueError(
            f"'{key}' in the configuration must be "
            f"one of {', '.join(valid_values)}."
        )
    return key == 'true'


from argparse import (
    ArgumentParser,
    RawTextHelpFormatter,
    Namespace,
    _SubParsersAction
)

from boomba.cli.description import ERROR


class BaseCommand:
    _description: str
    _epilog: str
    _parser: ArgumentParser
    _subparser: _SubParsersAction
    _args: Namespace
    
    # base command
    command: str = 'command'
    
    def __init__(self):
        self._create_parser()
        self._create_subparser()
        self.add_all()
        self._args = self.parse()
    
    def _create_parser(self) -> None:
        self._parser = ArgumentParser(
            description=self._description,
            epilog=self._epilog,
            formatter_class=RawTextHelpFormatter
        )
        
    def _create_subparser(self) -> None:
        self._subparser = self._parser.add_subparsers(
            dest='command',
            help='Available commands'
        )

    def add_all(self) -> None:
        '''
        Adds all the required arguments at once to the parser.\n
        This method acts like an abstract method.
        
        Example
        -------
        ```python
        # must implement this method as shown below
        def add_all(self) -> None:
            self.parser.add_argument('--name')
            self.parser.add_argument('--age', type=int)
        ```
        '''
        raise NotImplementedError(
            "Subclasses must implement this method to define arguments."
        )

    def parse(self):
        return self._parser.parse_args()

    def get_command(self) -> str:
        if self._args is None:
            print(ERROR['empty_args'])
        return getattr(self._args, self.command, None)
    
    def get_option(self, key: str) -> str:
        value = getattr(self._args, key.replace('-', ''), None)
        if value is None:
            print(ERROR['empty_args'])
        return value
    
    def help(self):
        self._parser.print_help()